/**
 * $Id: VolatileMemory.java 62 2010-06-29 22:06:12Z mviara $
 */
package j51.intel;


import j51.util.Hex;
import j51.util.Logger;
import j51.util.FastArray;

/**
 * Volatile implementation of memory.
 * 
 * @author Mario Viara
 * @version 1.00
 * 
 * @since 1.04
 */
public class VolatileMemory implements Memory
{
	private static Logger log = Logger.getLogger(VolatileMemory.class);
	private byte buffer[] = new byte[0];
	private String name;
	private FastArray mrs[];
	private FastArray mws[];
	private boolean readBusy[];
	private boolean writeBusy[];
	private boolean present[] = new boolean[0];
	private boolean writeListener;
	
	public VolatileMemory()
	{
		this(0);
	}
	
	public VolatileMemory(int size)
	{
		this("Memory",size);
	}

	public VolatileMemory(String name,int size)
	{
		this(name,"memory",size);
	}

	public VolatileMemory(String name,String suffix,int size)
	{
		this.name = name+"."+size+"."+suffix;
		setLimit(0x10000);
		setSize(size);

		writeListener = true;
	}

	public boolean getWriteListener()
	{
		return writeListener;
	}
	
	public void setWriteListener(boolean mode)
	{
		writeListener = mode;
	}
	
	public String getName()
	{
		return name;
	}

	private void setLimit(int limit)
	{
		if (limit < present.length)
			return;
		
		present = new boolean[limit];
		mrs = new FastArray[limit];
		mws = new FastArray[limit];
		readBusy = new boolean[limit];
		writeBusy = new boolean[limit];

		for (int i = 0 ; i < present.length ; i++)
		{
			readBusy[i] = false;
			writeBusy[i] = false;
			present[i] = false;
			mws[i] = new FastArray<MemoryWriteListener>();
			mrs[i] = new FastArray<MemoryReadListener>();
		}
	}
	
	public void setSize(int size)
	{
		setLimit(size);
		if (buffer.length != size)
		{
			setPresent(0,size);
			log.fine(name+" Resize from "+buffer.length+" to "+size);
			buffer = new byte[size];
			for (int i = 0 ; i < buffer.length ; i++)
				buffer[i] = -1;
		}
	}

	public int getSize()
	{
		return buffer.length;
	}


	public void setPresent(int from,int start)
	{
		for (int i = from ; i < start ; i++)
		     present[i] = true;
	}
	
	public boolean isPresent(int address)
	{
		return present[address];
	}

	public int readDirect(int address)
	{
		if (!isPresent(address))
			throw new java.lang.Error(name+" Read 0x"+Hex.bin2word(address)+" out of range, max 0x"+Hex.bin2word(buffer.length));

		if (address < buffer.length)
			return buffer[address] & 0xff;
		else
			return 0;
	}

	public int read(int address)
	{

		int value = readDirect(address);

		if (!readBusy[address])
		{
			readBusy[address] = true;
			
			
			FastArray<MemoryReadListener> mr = (FastArray<MemoryReadListener>) mrs[address];
			
			for (int i =  mr.size(); --i >= 0; )
				value = mr.get(i).readMemory(address,value);

			
			readBusy[address] = false;
		}
		else
		{
			log.fine("Read busy at "+Hex.bin2word(address)+ " Memory "+this);

		}
		
		return value & 0xff;
	}

	public void writeDirect(int address,int newValue)
	{
		if (!isPresent(address))
			throw new java.lang.Error(name+" Write 0x"+Hex.bin2word(address)+" out of range, max 0x"+Hex.bin2word(buffer.length));

		if (address < buffer.length)
		{
			buffer[address] = (byte)newValue;
		}

	}
	public void write(int address,int newValue)
	{
		
		int oldValue = readDirect(address);
		writeDirect(address,newValue);
		
		if (writeListener)
		{
			if (!writeBusy[address])
			{
				writeBusy[address] = true;

				FastArray<MemoryWriteListener> mw = (FastArray<MemoryWriteListener>) mws[address];
				
				for (int i = mw.size(); --i >= 0 ; )
					mw.get(i).writeMemory(address,newValue,oldValue);


				writeBusy[address] = false;
			}
			else
			{
				log.fine("Write busy at "+Hex.bin2word(address)+ " Memory "+this);
			}


		}

	}

	public void addMemoryReadListener(int address,MemoryReadListener l)
	{
		present[address] = true;
		mrs[address].add(l);
	}
	
	public void addMemoryWriteListener(int address,MemoryWriteListener l)
	{
		present[address] = true;
		mws[address].add(l);
	}

	public String toString()
	{
		return name;
	}

	
}
